@@ -28,10 +28,14 @@
 #endif
 
 #include <stdint.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include <time.h>
 #include <signal.h>
 #include <pthread.h>
 #include <sys/types.h>
+#include <unistd.h>
 
 #include "nwipe.h"
 #include "context.h"
@@ -42,6 +46,7 @@
 #include "logging.h"
 #include "gui.h"
 #include "temperature.h"
+#include "miscellaneous.h"
 
 #include <sys/ioctl.h> /* FIXME: Twice Included */
 #include <sys/shm.h>
@@ -49,12 +54,31 @@
 
 #include <parted/parted.h>
 #include <parted/debug.h>
+#include "conf.h"
 #include "version.h"
+#include "hpa_dco.h"
+#include "conf.h"
+#include <libconfig.h>
 
 int terminate_signal;
 int user_abort;
 int global_wipe_status;
 
+/* helper function for sorting */
+int devnamecmp( const void* a, const void* b )
+{
+    // nwipe_log( NWIPE_LOG_DEBUG, "a: %s, b: %s", ( *( nwipe_context_t** ) a)->device_name, ( *( nwipe_context_t** )
+    // b)->device_name );
+
+    int ldiff = strlen( ( *(nwipe_context_t**) a )->device_name ) - strlen( ( *(nwipe_context_t**) b )->device_name );
+    if( ldiff != 0 )
+    {
+        return ldiff;
+    }
+    int ret = strcmp( ( *(nwipe_context_t**) a )->device_name, ( *(nwipe_context_t**) b )->device_name );
+    return ( ret );
+}
+
 int main( int argc, char** argv )
 {
     int nwipe_optind;  // The result of nwipe_options().
@@ -64,6 +88,7 @@ int main( int argc, char** argv )
     int any_threads_still_running;  // used in wipe thread cancellation wait loop
     int thread_timeout_counter;  // timeout thread cancellation after THREAD_CANCELLATION_TIMEOUT seconds
     pthread_t nwipe_gui_thread = 0;  // The thread ID of the GUI thread.
+    pthread_t nwipe_temperature_thread = 0;  // The thread ID of the temperature update thread
     pthread_t nwipe_sigint_thread;  // The thread ID of the sigint handler.
 
     char modprobe_command[] = "modprobe %s";
@@ -82,12 +107,6 @@ int main( int argc, char** argv )
     /* The generic result buffer. */
     int r;
 
-    /* Log nwipes version */
-    nwipe_log( NWIPE_LOG_INFO, "%s", banner );
-
-    /* Log OS info */
-    nwipe_log_OSinfo();
-
     /* Initialise the termintaion signal, 1=terminate nwipe */
     terminate_signal = 0;
 
@@ -112,11 +131,67 @@ int main( int argc, char** argv )
     /* Initialised and populated in device scan.     */
     nwipe_context_t** c1 = 0;
 
+    if( geteuid() != 0 )
+    {
+        printf( "nwipe must run with root permissions, which is not the case.\nAborting\n" );
+        exit( 99 );
+    }
+
     int wipe_threads_started = 0;
 
-    /* Parse command line options. */
+    /** NOTE ** NOTE ** NOTE ** NOTE ** NOTE ** NOTE ** NOTE ** NOTE ** NOTE **
+     * Important Note: if you want nwipe_log messages to go into the logfile
+     * any 'nwipe_log()' commands must appear after the options are parsed here,
+     * else they will appear in the console but not in the logfile, that is,
+     * assuming you specified a log file on the command line as an nwipe option.
+     */
+
+    /*****************************
+     * Parse command line options.
+     */
+
+    /* Initialise the libconfig code that handles nwipe.conf */
+    nwipe_conf_init();
+
     nwipe_optind = nwipe_options_parse( argc, argv );
 
+    /* Log nwipes version */
+    nwipe_log( NWIPE_LOG_INFO, "%s", banner );
+
+    /* Log OS info */
+    nwipe_log_OSinfo();
+
+    /* Check that hdparm exists, we use hdparm for some HPA/DCO detection etc, if not
+     * exit nwipe. These checks are required if the PATH environment is not setup !
+     * Example: Debian sid 'su' as opposed to 'su -'
+     */
+    if( system( "which hdparm > /dev/null 2>&1" ) )
+    {
+        if( system( "which /sbin/hdparm > /dev/null 2>&1" ) )
+        {
+            if( system( "which /usr/bin/hdparm > /dev/null 2>&1" ) )
+            {
+                nwipe_log( NWIPE_LOG_WARNING, "hdparm command not found." );
+                nwipe_log( NWIPE_LOG_WARNING,
+                           "Required by nwipe for HPA/DCO detection & correction and ATA secure erase." );
+                nwipe_log( NWIPE_LOG_WARNING, "** Please install hdparm **\n" );
+                cleanup();
+                exit( 1 );
+            }
+        }
+    }
+
+    /* Check if the given path for PDF reports is a writeable directory */
+    if( strcmp( nwipe_options.PDFreportpath, "noPDF" ) != 0 )
+    {
+        if( access( nwipe_options.PDFreportpath, W_OK ) != 0 )
+        {
+            nwipe_log( NWIPE_LOG_ERROR, "PDFreportpath %s is not a writeable directory.", nwipe_options.PDFreportpath );
+            cleanup();
+            exit( 2 );
+        }
+    }
+
     if( nwipe_optind == argc )
     {
         /* File names were not given by the user.  Scan for devices. */
@@ -148,13 +223,18 @@ int main( int argc, char** argv )
         nwipe_enumerated = nwipe_device_get( &c1, argv, argc );
         if( nwipe_enumerated == 0 )
         {
-            nwipe_log( NWIPE_LOG_ERROR, "Devices not found. Check you're not excluding drives unnecessarily." );
-            printf( "No drives found\n" );
+            nwipe_log( NWIPE_LOG_ERROR, "Devices not found. Check you're not excluding drives unnecessarily," );
+            nwipe_log( NWIPE_LOG_ERROR, "and you are running nwipe as sudo or as root." );
+            printf( "Devices not found, check you're not excluding drives unnecessarily \n and you are running nwipe "
+                    "as sudo or as root." );
             cleanup();
             exit( 1 );
         }
     }
 
+    /* sort list of devices here */
+    qsort( (void*) c1, (size_t) nwipe_enumerated, sizeof( nwipe_context_t* ), devnamecmp );
+
     if( terminate_signal == 1 )
     {
         cleanup();
@@ -166,6 +246,12 @@ int main( int argc, char** argv )
 
     /* The array of pointers to contexts that will actually be wiped. */
     nwipe_context_t** c2 = (nwipe_context_t**) malloc( nwipe_enumerated * sizeof( nwipe_context_t* ) );
+    if( c2 == NULL )
+    {
+        nwipe_log( NWIPE_LOG_ERROR, "memory allocation for c2 failed" );
+        cleanup();
+        exit( 1 );
+    }
 
     /* Open the entropy source. */
     nwipe_entropy = open( NWIPE_KNOB_ENTROPY, O_RDONLY );
@@ -255,7 +341,7 @@ int main( int argc, char** argv )
     /* load the drivetemp module */
     if( system( final_cmd_modprobe ) != 0 )
     {
-        nwipe_log( NWIPE_LOG_WARNING, "hwmon: Unable to load module drivetemp, temperatures may be unavilable." );
+        nwipe_log( NWIPE_LOG_WARNING, "hwmon: Unable to load module drivetemp, temperatures may be unavailable." );
         nwipe_log( NWIPE_LOG_WARNING, "hwmon: It's possible the drivetemp software isn't modular but built-in" );
         nwipe_log( NWIPE_LOG_WARNING, "hwmon: to the kernel, as is the case with ShredOS.x86_64 in which case" );
         nwipe_log( NWIPE_LOG_WARNING, "hwmon: the temperatures will actually be available despite this issue." );
@@ -294,7 +380,13 @@ int main( int argc, char** argv )
             nwipe_log( NWIPE_LOG_NOTICE, "hwmon: Device %s hwmon path = %s", c1[i]->device_name, c1[i]->temp1_path );
         }
 
-        nwipe_update_temperature( c1[i] );
+        // nwipe_update_temperature( c1[i] );
+
+        /* Log the temperature crtical, highest, lowest and lowest critical temperature
+         * limits to nwipes log file using the INFO catagory
+         */
+
+        nwipe_log_drives_temperature_limits( c1[i] );
     }
 
     /* Check for initialization errors. */
@@ -305,6 +397,15 @@ int main( int argc, char** argv )
         return -1;
     }
 
+    /* Set up the data structures to pass the temperature thread the data it needs */
+    nwipe_thread_data_ptr_t nwipe_temperature_thread_data;
+    nwipe_temperature_thread_data.c = c1;
+    nwipe_temperature_thread_data.nwipe_misc_thread_data = &nwipe_misc_thread_data;
+
+    /* Fork the temperature thread */
+    errno = pthread_create(
+        &nwipe_temperature_thread, NULL, nwipe_update_temperature_thread, &nwipe_temperature_thread_data );
+
     /* Start the ncurses interface. */
     if( !nwipe_options.nogui )
         nwipe_gui_init();
@@ -326,10 +427,17 @@ int main( int argc, char** argv )
         }
         else
         {
+            if( nwipe_options.PDF_preview_details == 1 )
+            {
+                nwipe_gui_preview_org_customer( SHOWING_PRIOR_TO_DRIVE_SELECTION );
+            }
+
             nwipe_gui_select( nwipe_enumerated, c1 );
         }
     }
 
+    /* Initialise some of the variables in the drive contexts
+     */
     for( i = 0; i < nwipe_enumerated; i++ )
     {
         /* Set the PRNG implementation, which must always come after the function nwipe_gui_select ! */
@@ -346,6 +454,9 @@ int main( int argc, char** argv )
 
         /* Initialise the wipe result value */
         c1[i]->result = 0;
+
+        /* Initialise the variable that tracks how much of the drive has been erased */
+        c1[i]->bytes_erased = 0;
     }
 
     /* Pass the number selected to the struct for other threads */
@@ -431,7 +542,8 @@ int main( int argc, char** argv )
                 nwipe_log( NWIPE_LOG_NOTICE, "%s has serial number %s", c2[i]->device_name, c2[i]->device_serial_no );
             }
 
-            /* Do sector size and block size checking. */
+            /* Do sector size and block size checking. I don't think this does anything useful as logical/Physical
+             * sector sizes are obtained by libparted in check.c */
             if( ioctl( c2[i]->device_fd, BLKSSZGET, &c2[i]->device_sector_size ) == 0 )
             {
 
@@ -702,6 +814,42 @@ int main( int argc, char** argv )
         thread_timeout_counter--;
         sleep( 1 );
     }
+
+    /* Now all the wipe threads have finished, we can issue a terminate_signal = 1
+     * which will cause the temperature update thread to terminate, this is necessary
+     * because in gui mode the terminate_signal is set when the user presses a key to
+     * exit on completion of all the wipes, however in non gui mode that code isn't
+     * active (being in the gui section) so here we need to set the terminate signal
+     * specifically for a completed wipes/s just for non gui mode.
+     */
+    terminate_signal = 1;
+
+    /* Kill the temperature update thread */
+    if( nwipe_temperature_thread )
+    {
+        if( nwipe_options.verbose )
+        {
+            nwipe_log( NWIPE_LOG_INFO, "Cancelling the temperature thread." );
+        }
+
+        /* We don't want to use pthread_cancel as our temperature thread is aware of the control-c
+         *  signal and will exit itself we just join the temperature thread and wait for confirmation
+         */
+        r = pthread_join( nwipe_temperature_thread, NULL );
+        if( r != 0 )
+        {
+            nwipe_log( NWIPE_LOG_WARNING,
+                       "main()>pthread_join():Error when waiting for temperature thread to cancel." );
+        }
+        else
+        {
+            if( nwipe_options.verbose )
+            {
+                nwipe_log( NWIPE_LOG_INFO, "temperature thread has been cancelled" );
+            }
+        }
+    }
+
     if( nwipe_options.verbose )
     {
         for( i = 0; i < nwipe_selected; i++ )
@@ -936,11 +1084,12 @@ void* signal_hand( void* ptr )
 int cleanup()
 {
     int i;
-    extern int log_elements_displayed;
-    extern int log_elements_allocated;
+    extern int log_elements_displayed;  // initialised and found in logging.c
+    extern int log_elements_allocated;  // initialised and found in logging.c
     extern char** log_lines;
+    extern config_t nwipe_cfg;
 
-    /* Print the logs held in memory. */
+    /* Print the logs held in memory to the console */
     for( i = log_elements_displayed; i < log_elements_allocated; i++ )
     {
         printf( "%s\n", log_lines[i] );
@@ -958,7 +1107,10 @@ int cleanup()
         free( log_lines );
     }
 
-    /* TODO: All other cleanup required */
+    /* Deallocate libconfig resources */
+    config_destroy( &nwipe_cfg );
+
+    /* TODO: Any other cleanup required ? */
 
     return 0;
 }
